@@ -1,4 +1,21 @@
-# -*- coding: utf-8 -*-
+import errno
+import glob
+import hashlib
+import json
+import os
+import shutil
+import sys
+import tempfile
+import zipfile
+from datetime import datetime
+from urllib.parse import urlsplit
+
+try:
+    import requests
+
+    requests_installed = True
+except ImportError:
+    requests_installed = False
 
 from fsutil.metadata import (
     __author__,
@@ -10,32 +27,6 @@ from fsutil.metadata import (
     __version__,
 )
 
-import datetime as dt
-import errno
-import glob
-import hashlib
-import json
-import os
-
-try:
-    import requests
-
-    requests_installed = True
-except ImportError:
-    requests_installed = False
-
-import shutil
-import sys
-import zipfile
-
-try:
-    # python 2
-    from urlparse import urlsplit
-except ImportError:
-    # python 3
-    from urllib.parse import urlsplit
-
-
 __all__ = [
     "__author__",
     "__copyright__",
@@ -121,7 +112,6 @@ __all__ = [
     "write_file_json",
 ]
 
-PY2 = bool(sys.version_info.major == 2)
 SIZE_UNITS = ["bytes", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"]
 
 
@@ -138,7 +128,7 @@ def assert_dir(path):
     Raise an OSError if the given path doesn't exist or it is not a directory.
     """
     if not is_dir(path):
-        raise OSError("Invalid directory path: {}".format(path))
+        raise OSError(f"Invalid directory path: {path}")
 
 
 def assert_exists(path):
@@ -146,7 +136,7 @@ def assert_exists(path):
     Raise an OSError if the given path doesn't exist.
     """
     if not exists(path):
-        raise OSError("Invalid item path: {}".format(path))
+        raise OSError(f"Invalid item path: {path}")
 
 
 def assert_file(path):
@@ -154,7 +144,7 @@ def assert_file(path):
     Raise an OSError if the given path doesn't exist or it is not a file.
     """
     if not is_file(path):
-        raise OSError("Invalid file path: {}".format(path))
+        raise OSError(f"Invalid file path: {path}")
 
 
 def assert_not_dir(path):
@@ -162,7 +152,7 @@ def assert_not_dir(path):
     Raise an OSError if the given path is an existing directory.
     """
     if is_dir(path):
-        raise OSError("Invalid path, directory already exists: {}".format(path))
+        raise OSError(f"Invalid path, directory already exists: {path}")
 
 
 def assert_not_exists(path):
@@ -170,7 +160,7 @@ def assert_not_exists(path):
     Raise an OSError if the given path already exists.
     """
     if exists(path):
-        raise OSError("Invalid path, item already exists: {}".format(path))
+        raise OSError(f"Invalid path, item already exists: {path}")
 
 
 def assert_not_file(path):
@@ -178,7 +168,7 @@ def assert_not_file(path):
     Raise an OSError if the given path is an existing file.
     """
     if is_file(path):
-        raise OSError("Invalid path, file already exists: {}".format(path))
+        raise OSError(f"Invalid path, file already exists: {path}")
 
 
 def _clean_dir_empty_dirs(path):
@@ -218,9 +208,10 @@ def convert_size_bytes_to_string(size):
     while (size >= 1024) and (factor <= factor_limit):
         size /= 1024
         factor += 1
-    s_format = "{:.2f} {}" if (factor > 1) else "{:.0f} {}"
-    s = s_format.format(size, units[factor])
-    return s
+    size_units = units[factor]
+    size_str = f"{size:.2f}" if (factor > 1) else f"{size:.0f}"
+    size_str = f"{size_str} {size_units}"
+    return size_str
 
 
 def convert_size_string_to_bytes(size):
@@ -373,7 +364,7 @@ def delete_files(*paths):
     remove_files(*paths)
 
 
-def download_file(url, dirpath, filename=None, chunk_size=8192, **kwargs):
+def download_file(url, dirpath=None, filename=None, chunk_size=8192, **kwargs):
     """
     Download a file from url to dirpath.
     If filename is provided, the file will be named using filename.
@@ -381,6 +372,7 @@ def download_file(url, dirpath, filename=None, chunk_size=8192, **kwargs):
     """
     _require_requests_installed()
     # https://stackoverflow.com/a/16696317/2096218
+    dirpath = dirpath or tempfile.gettempdir()
     filename = filename or get_filename(url) or "download"
     filepath = join_path(dirpath, filename)
     make_dirs_for_file(filepath)
@@ -435,7 +427,7 @@ def get_dir_creation_date(path):
     """
     assert_dir(path)
     creation_timestamp = os.path.getctime(path)
-    creation_date = dt.datetime.fromtimestamp(creation_timestamp)
+    creation_date = datetime.fromtimestamp(creation_timestamp)
     return creation_date
 
 
@@ -464,7 +456,7 @@ def get_dir_last_modified_date(path):
             last_modified_timestamp = max(
                 last_modified_timestamp, os.path.getmtime(filepath)
             )
-    last_modified_date = dt.datetime.fromtimestamp(last_modified_timestamp)
+    last_modified_date = datetime.fromtimestamp(last_modified_timestamp)
     return last_modified_date
 
 
@@ -513,7 +505,7 @@ def get_file_creation_date(path):
     """
     assert_file(path)
     creation_timestamp = os.path.getctime(path)
-    creation_date = dt.datetime.fromtimestamp(creation_timestamp)
+    creation_date = datetime.fromtimestamp(creation_timestamp)
     return creation_date
 
 
@@ -553,7 +545,7 @@ def get_file_last_modified_date(path):
     """
     assert_file(path)
     last_modified_timestamp = os.path.getmtime(path)
-    last_modified_date = dt.datetime.fromtimestamp(last_modified_timestamp)
+    last_modified_date = datetime.fromtimestamp(last_modified_timestamp)
     return last_modified_date
 
 
@@ -645,7 +637,7 @@ def join_filename(basename, extension):
     """
     basename = basename.rstrip(".").strip()
     extension = extension.replace(".", "").strip()
-    filename = "{}.{}".format(basename, extension)
+    filename = f"{basename}.{extension}"
     return filename
 
 
@@ -683,16 +675,6 @@ def list_files(path):
     return _filter_paths(path, os.listdir(path), predicate=is_file)
 
 
-def _make_dirs_py2(path):
-    # https://stackoverflow.com/questions/12517451/automatically-creating-directories-with-file-output
-    try:
-        os.makedirs(path)
-    except OSError as e:
-        # Guard against race condition
-        if e.errno != errno.EEXIST:
-            raise e
-
-
 def make_dirs(path):
     """
     Create the directories needed to ensure that the given path exists.
@@ -701,9 +683,6 @@ def make_dirs(path):
     if is_dir(path):
         return
     assert_not_file(path)
-    if PY2:
-        _make_dirs_py2(path)
-        return
     os.makedirs(path, exist_ok=True)
 
 
@@ -757,8 +736,7 @@ def read_file(path, encoding="utf-8"):
     """
     assert_file(path)
     content = ""
-    options = {} if PY2 else {"encoding": encoding}
-    with open(path, "r", **options) as file:
+    with open(path, "r", encoding=encoding) as file:
         content = file.read()
     return content
 
@@ -807,9 +785,9 @@ def _read_file_lines_in_range(path, line_start=0, line_end=-1, encoding="utf-8")
         lines_count = read_file_lines_count(path)
         # normalize negative indexes
         if line_start_negative:
-            line_start += lines_count
+            line_start = max(0, line_start + lines_count)
         if line_end_negative:
-            line_end += lines_count
+            line_end = min(line_end + lines_count, lines_count - 1)
     with open(path, "rb") as file:
         file.seek(0)
         line_index = 0
@@ -956,8 +934,7 @@ def _search_paths(path, pattern):
     """
     assert_dir(path)
     pathname = os.path.join(path, pattern)
-    options = {} if PY2 else {"recursive": True}
-    paths = glob.glob(pathname, **options)
+    paths = glob.glob(pathname, recursive=True)
     return paths
 
 
@@ -1010,8 +987,7 @@ def write_file(path, content, append=False, encoding="utf-8"):
     """
     make_dirs_for_file(path)
     mode = "a" if append else "w"
-    options = {} if PY2 else {"encoding": encoding}
-    with open(path, mode, **options) as file:
+    with open(path, mode, encoding=encoding) as file:
         file.write(content)
 
 
@@ -1031,6 +1007,14 @@ def write_file_json(
     """
     Write a json file at the given path with the specified data encoded in json format.
     """
+
+    def default_encoder(obj):
+        if isinstance(obj, datetime):
+            return obj.isoformat()
+        elif isinstance(obj, set):
+            return list(obj)
+        return str(obj)
+
     content = json.dumps(
         data,
         skipkeys=skipkeys,
@@ -1040,7 +1024,7 @@ def write_file_json(
         cls=cls,
         indent=indent,
         separators=separators,
-        default=default,
+        default=default or default_encoder,
         sort_keys=sort_keys,
     )
     write_file(path, content)
